from bmsdna.lakeapi.core.config import Configs
from bmsdna.lakeapi.api.api import init_lakeapi
from bmsdna.lakeapi.core.config import get_default_config, BasicConfig
from bmsdna.lakeapi.core.types import MetadataDetailResult, MetadataSchemaField, MetadataSchemaFieldType
from bmsdna.lakeapi.core.uservalidation import add_user_middlware

import sys

sys.setrecursionlimit(10_000)
